@@ -1,5 +1,4 @@
 #!/usr/bin/env python
-# -*- coding: utf-8 -*-
 
 """
 Implementation of gpustat
@@ -8,11 +7,7 @@ Implementation of gpustat
 @url https://github.com/wookayin/gpustat
 """
 
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-from __future__ import unicode_literals
-
+from typing import Sequence
 import json
 import locale
 import os.path
@@ -20,13 +15,13 @@ import platform
 import sys
 import time
 from datetime import datetime
+from io import StringIO
 
-from six.moves import cStringIO as StringIO
 import psutil
 from blessed import Terminal
 
-from gpustat.nvml import pynvml as N
 import gpustat.util as util
+from gpustat.nvml import pynvml as N
 
 
 NOT_SUPPORTED = 'Not Supported'
@@ -55,6 +50,10 @@ class GPUStat(object):
     def __getitem__(self, key):
         return self.entry[key]
 
+    @property
+    def available(self):
+        return True
+
     @property
     def index(self):
         """
@@ -177,16 +176,18 @@ class GPUStat(object):
         """
         return self.entry['processes']
 
-    def print_to(self, fp,
+    def print_to(self, fp, *,
                  with_colors=True,    # deprecated arg
                  show_cmd=False,
                  show_full_cmd=False,
+                 no_processes=False,
                  show_user=False,
                  show_pid=False,
                  show_fan_speed=None,
                  show_codec="",
                  show_power=None,
                  gpuname_width=None,
+                 eol_char=os.linesep,
                  term=None,
                  ):
         if term is None:
@@ -207,13 +208,16 @@ class GPUStat(object):
         colors['C0'] = term.normal
         colors['C1'] = term.cyan
         colors['CBold'] = term.bold
-        colors['CName'] = term.blue
+        colors['CName'] = _conditional(lambda: self.available,
+                                       term.blue, term.red)
         colors['CTemp'] = _conditional(lambda: self.temperature < 50,
                                        term.red, term.bold_red)
         colors['FSpeed'] = _conditional(lambda: self.fan_speed < 30,
                                         term.cyan, term.bold_cyan)
-        colors['CMemU'] = term.bold_yellow
-        colors['CMemT'] = term.yellow
+        colors['CMemU'] = _conditional(lambda: self.available,
+                                       term.bold_yellow, term.bold_black)
+        colors['CMemT'] = _conditional(lambda: self.available,
+                                       term.yellow, term.bold_black)
         colors['CMemP'] = term.yellow
         colors['CCPUMemU'] = term.yellow
         colors['CUser'] = term.bold_black   # gray
@@ -277,13 +281,23 @@ class GPUStat(object):
         reps += " | %(C1)s%(CMemU)s{entry[memory.used]:>5}%(C0)s " \
             "/ %(CMemT)s{entry[memory.total]:>5}%(C0)s MB"
         reps = (reps) % colors
+
+        class entry_repr_accessor:
+            def __init__(self, entry):
+                self.entry = entry
+            def __getitem__(self, key):
+                return _repr(self.entry[key])
+
         reps = reps.format(
-            entry={k: _repr(v) for k, v in self.entry.items()},
+            entry=entry_repr_accessor(self.entry),
             entry_name=util.shorten_left(
                 self.entry["name"], width=gpuname_width, placeholder='…'),
             gpuname_width=gpuname_width or DEFAULT_GPUNAME_WIDTH
         )
-        reps += " |"
+
+        # Add " |" only if processes information is to be added.
+        if not no_processes:
+            reps += " |"
 
         def process_repr(p):
             r = ''
@@ -323,14 +337,14 @@ class GPUStat(object):
 
         processes = self.entry['processes']
         full_processes = []
-        if processes is None:
+        if processes is None and not no_processes:
             # None (not available)
             reps += ' ({})'.format(NOT_SUPPORTED)
-        else:
+        elif not no_processes:
             for p in processes:
                 reps += ' ' + process_repr(p)
                 if show_full_cmd:
-                    full_processes.append(os.linesep + full_process_info(p))
+                    full_processes.append(eol_char + full_process_info(p))
         if show_full_cmd and full_processes:
             full_processes[-1] = full_processes[-1].replace('├', '└', 1)
             reps += ''.join(full_processes)
@@ -345,7 +359,25 @@ class GPUStat(object):
         return o
 
 
-class GPUStatCollection(object):
+class InvalidGPU(GPUStat):
+    class FallbackDict(dict):
+        def __missing__(self, key):
+            return "?"
+
+    def __init__(self, gpu_index, message, ex):
+        super().__init__(self.FallbackDict(
+            index=gpu_index,
+            name=message,
+            processes=None
+        ))
+        self.exception = ex
+
+    @property
+    def available(self):
+        return False
+
+
+class GPUStatCollection(Sequence[GPUStat]):
 
     global_processes = {}
 
@@ -364,7 +396,7 @@ class GPUStatCollection(object):
                 del GPUStatCollection.global_processes[pid]
 
     @staticmethod
-    def new_query(debug=False):
+    def new_query(debug=False, id=None) -> 'GPUStatCollection':
         """Query the information of all the GPUs on local machine"""
 
         N.nvmlInit()
@@ -378,19 +410,27 @@ class GPUStatCollection(object):
         def get_gpu_info(handle):
             """Get one GPU information specified by nvml handle"""
 
+            def safepcall(fn, error_value):
+                # Ignore the exception from psutil when the process is gone
+                # at the moment of querying. See #144.
+                return util.safecall(
+                    fn, error_value=error_value,
+                    exc_types=(psutil.AccessDenied, psutil.NoSuchProcess,
+                               FileNotFoundError))
+
             def get_process_info(nv_process):
                 """Get the process information of specific pid"""
                 process = {}
                 if nv_process.pid not in GPUStatCollection.global_processes:
                     GPUStatCollection.global_processes[nv_process.pid] = \
                         psutil.Process(pid=nv_process.pid)
-                ps_process = GPUStatCollection.global_processes[nv_process.pid]
+                ps_process: psutil.Process = GPUStatCollection.global_processes[nv_process.pid]
 
                 # TODO: ps_process is being cached, but the dict below is not.
-                process['username'] = ps_process.username()
+                process['username'] = safepcall(ps_process.username, '?')
                 # cmdline returns full path;
                 # as in `ps -o comm`, get short cmdnames.
-                _cmdline = ps_process.cmdline()
+                _cmdline = safepcall(ps_process.cmdline, [])
                 if not _cmdline:
                     # sometimes, zombie or unknown (e.g. [kworker/8:2H])
                     process['command'] = '?'
@@ -403,10 +443,13 @@ class GPUStatCollection(object):
                 usedmem = nv_process.usedGpuMemory // MB if \
                           nv_process.usedGpuMemory else None
                 process['gpu_memory_usage'] = usedmem
-                process['cpu_percent'] = ps_process.cpu_percent()
-                process['cpu_memory_usage'] = \
-                    round((ps_process.memory_percent() / 100.0) *
-                          psutil.virtual_memory().total)
+
+                process['cpu_percent'] = safepcall(ps_process.cpu_percent, 0.0)
+                process['cpu_memory_usage'] = safepcall(
+                    lambda: round((ps_process.memory_percent() / 100.0) *
+                                  psutil.virtual_memory().total),
+                    0.0)
+
                 process['pid'] = nv_process.pid
                 return process
 
@@ -428,7 +471,9 @@ class GPUStatCollection(object):
                 fan_speed = None  # Not supported
 
             try:
-                memory = N.nvmlDeviceGetMemoryInfo(handle)  # in Bytes
+                # memory: in Bytes
+                # Note that this is a compat-patched API (see gpustat.nvml)
+                memory = N.nvmlDeviceGetMemoryInfo(handle)
             except N.NVMLError as e:
                 log.add_exception("memory", e)
                 memory = None  # Not supported
@@ -448,7 +493,7 @@ class GPUStatCollection(object):
             try:
                 utilization_dec = N.nvmlDeviceGetDecoderUtilization(handle)
             except N.NVMLError as e:
-                log.add_exception("utilization_dnc", e)
+                log.add_exception("utilization_dec", e)
                 utilization_dec = None  # Not supported
 
             try:
@@ -510,8 +555,8 @@ class GPUStatCollection(object):
                 time.sleep(0.1)
                 for process in processes:
                     pid = process['pid']
-                    cache_process = GPUStatCollection.global_processes[pid]
-                    process['cpu_percent'] = cache_process.cpu_percent()
+                    cache_process: psutil.Process = GPUStatCollection.global_processes[pid]
+                    process['cpu_percent'] = safepcall(cache_process.cpu_percent, 0)
 
             index = N.nvmlDeviceGetIndex(handle)
 
@@ -543,10 +588,27 @@ class GPUStatCollection(object):
         gpu_list = []
         device_count = N.nvmlDeviceGetCount()
 
-        for index in range(device_count):
-            handle = N.nvmlDeviceGetHandleByIndex(index)
-            gpu_info = get_gpu_info(handle)
-            gpu_stat = GPUStat(gpu_info)
+        if id is None:
+            gpus_to_query = range(device_count)
+        elif isinstance(id, str):
+            gpus_to_query = [int(i) for i in id.split(',')]
+        elif isinstance(id, Sequence):
+            gpus_to_query = [int(i) for i in id]
+        else:
+            raise TypeError(f"Unknown id: {id}")
+
+        for index in gpus_to_query:
+            try:
+                handle = N.nvmlDeviceGetHandleByIndex(index)
+                gpu_info = get_gpu_info(handle)
+                gpu_stat = GPUStat(gpu_info)
+            except N.NVMLError_Unknown as e:
+                gpu_stat = InvalidGPU(index, "((Unknown Error))", e)
+            except N.NVMLError_GpuIsLost as e:
+                gpu_stat = InvalidGPU(index, "((GPU is lost))", e)
+
+            if isinstance(gpu_stat, InvalidGPU):
+                log.add_exception("GPU %d" % index, gpu_stat.exception)
             gpu_list.append(gpu_stat)
 
         # 2. additional info (driver version, etc).
@@ -579,11 +641,13 @@ class GPUStatCollection(object):
 
     # --- Printing Functions ---
 
-    def print_formatted(self, fp=sys.stdout, force_color=False, no_color=False,
+    def print_formatted(self, fp=sys.stdout, *,
+                        force_color=False, no_color=False,
                         show_cmd=False, show_full_cmd=False, show_user=False,
                         show_pid=False, show_fan_speed=None,
                         show_codec="", show_power=None,
                         gpuname_width=None, show_header=True,
+                        no_processes=False,
                         eol_char=os.linesep,
                         ):
         # ANSI color configuration
@@ -604,7 +668,7 @@ class GPUStatCollection(object):
 
         # appearance settings
         if gpuname_width is None:
-            gpuname_width = max([len(g.entry['name']) for g in self])
+            gpuname_width = max([len(g.entry['name']) for g in self] + [0])
 
         # header
         if show_header:
@@ -635,15 +699,20 @@ class GPUStatCollection(object):
             g.print_to(fp,
                        show_cmd=show_cmd,
                        show_full_cmd=show_full_cmd,
+                       no_processes=no_processes,
                        show_user=show_user,
                        show_pid=show_pid,
                        show_fan_speed=show_fan_speed,
                        show_codec=show_codec,
                        show_power=show_power,
                        gpuname_width=gpuname_width,
+                       eol_char=eol_char,
                        term=t_color)
             fp.write(eol_char)
 
+        if len(self.gpus) == 0:
+            print(t_color.yellow("(No GPUs are available)"))
+
         fp.flush()
 
     def jsonify(self):
@@ -668,9 +737,28 @@ class GPUStatCollection(object):
         fp.flush()
 
 
-def new_query():
+def new_query() -> GPUStatCollection:
     '''
     Obtain a new GPUStatCollection instance by querying nvidia-smi
     to get the list of GPUs and running process information.
     '''
     return GPUStatCollection.new_query()
+
+
+def gpu_count() -> int:
+    '''Return the number of available GPUs in the system.'''
+    try:
+        N.nvmlInit()
+        return N.nvmlDeviceGetCount()
+    except N.NVMLError:
+        return 0  # fallback
+    finally:
+        try:
+            N.nvmlShutdown()
+        except N.NVMLError:
+            pass
+
+
+def is_available() -> bool:
+    '''Return True if the NVML library and GPU devices are available.'''
+    return gpu_count() > 0
